from base64 import b64decode, b64encode

import boto3


def encrypt(text, key_id='alias/kite_api_key'):
	kms = boto3.client('kms')
	return b64encode(kms.encrypt(KeyId=key_id, Plaintext=bytes(text, encoding='utf8'))['CiphertextBlob']).decode('utf8')


def decrypt(text, key_id='alias/kite_api_key'):
	kms = boto3.client('kms')
	return kms.decrypt(KeyId=key_id, CiphertextBlob=bytes(b64decode(text)))['Plaintext'].decode('utf8')


def get_from_s3(bucket, file_path):
	s3 = boto3.client('s3')
	file_obj = s3.get_object(Bucket=bucket, Key=file_path)
	
	return file_obj['Body'].read().decode('utf-8')


def upload_to_s3(bucket, file, key):
	s3 = boto3.client('s3')
	s3.put_object(Body=file, Bucket=bucket, Key=key)


def download_s3(bucket, s3_path, save_path):
	s3 = boto3.client('s3')
	s3.download_file(bucket, s3_path, save_path)



import json
import os
from hashlib import blake2b
from http.cookies import SimpleCookie

import jwt
from jwt.exceptions import PyJWTError

ACAO = 'Access-Control-Allow-Origin'


def validate_jwt(encoded_jwt, secret):
	try:
		return 200, jwt.decode(encoded_jwt, secret, algorithms='HS256')
	except PyJWTError as e:
		return 510, str(e)


def read_cookie(cookie, cookie_name):
	cookie = SimpleCookie(cookie)
	
	return cookie[cookie_name].value


def origin_verification(headers):
	allowed_origins = ['http://localhost:5173', 'https://stockemy.in', 'https://oms.stockemy.in']
	origin = headers.get('origin') or headers.get('Origin')
	
	if origin not in allowed_origins:
		code, message, origin = 510, 'Invalid origin', '*'
	else:
		code, message, origin = 200, 'Valid origin', origin
	
	return code, message, origin


def request_verification_flow(headers):
	code, message, origin = origin_verification(headers)
	if code != 200:
		return {'statusCode': code, 'body': json.dumps({'message': message}), 'headers': {ACAO: origin}}
	
	try:
		cookie = headers.get('cookie') or headers.get('Cookie')
		jwt_token = read_cookie(cookie, 'jwt_token')
	except KeyError:
		code, message = 510, 'Missing jwt_token in cookie'
		return {'statusCode': code, 'body': json.dumps({'message': message}), 'headers': {ACAO: origin}}
	
	code, message = validate_jwt(jwt_token, secret=os.environ['JWT_KEY'])
	if code != 200:
		return {'statusCode': code, 'body': json.dumps({'message': message}), 'headers': {ACAO: origin}}
	
	return message, origin


def get_hashed_password(password, key):
	h = blake2b(key=key.encode(), digest_size=16)
	h.update(password.encode())
	
	return h.hexdigest()



import boto3

dy = boto3.resource('dynamodb')


def replace_in_dynamo(table_name, item):
	table = dy.Table(table_name)
	table.put_item(Item=item)


def get_from_dynamo(table_name, key, proj_expr):
	table = dy.Table(table_name)
	res = table.get_item(Key=key, ProjectionExpression=proj_expr)
	
	return res.get('Item')


def update_in_dynamo(table_name, key, update_expr, expr_attr_vals, expr_attr_names=None):
	table = dy.Table(table_name)
	table.update_item(Key=key, UpdateExpression=update_expr, ExpressionAttributeValues=expr_attr_vals,
	                  ExpressionAttributeNames=expr_attr_names)


def delete_from_dynamo(table_name, key):
	table = dy.Table(table_name)
	table.delete_item(Key=key)


def query_dynamo(table_name, key, proj_expr, expr_attr_names=None):
	table = dy.Table(table_name)
	res = table.query(KeyConditionExpression=key, ProjectionExpression=proj_expr,
	                  ExpressionAttributeNames=expr_attr_names)
	
	return res.get('Items')


# print(query_dynamo('prod_watchlist', Key('user_id').eq('123'), 'list_id, #items', {'#items': 'items'}))



import json
import os
import re
from configparser import ConfigParser

import boto3
import requests as r
import urllib3
from urllib3.exceptions import InsecureRequestWarning

from quasi_utils.aws_utils import decrypt

urllib3.disable_warnings(InsecureRequestWarning)


def request(method, url, headers, data=None, params=None, _json=None, timeout=5, verify=False, allow_redirects=True,
            session=None):
	if session:
		res = session.request(method, url, data=data, headers=headers, json=_json, params=params, verify=verify,
		                      allow_redirects=allow_redirects, timeout=timeout, proxies=None)
	else:
		res = r.request(method, url, data=data, headers=headers, json=_json, params=params, verify=verify,
		                allow_redirects=allow_redirects, timeout=timeout, proxies=None)
	
	if res.status_code == 403:
		return {'status_code': 403}
	
	try:
		return json.loads(res.content.decode('utf8'))
	except json.decoder.JSONDecodeError:
		return res.content


def base(price, rounder=0.05, precision=2):
	return round(rounder * round(float(price) / rounder), precision)


def get_details(config_name, data_dir):
	config = ConfigParser()
	config.read(os.path.join(data_dir, 'config.ini'))
	z = config[config_name]
	api_key, api_secret, access_token = decrypt(z['api_key']), decrypt(z['api_secret']), decrypt(z['access_token'])
	
	return api_key, api_secret, access_token


def get_trade_number(orders):
	trade_num = 0
	
	if not orders:
		return trade_num
	
	for order in orders:
		if order['transaction_type'] == 'SELL':
			trade_num += 1
	
	return trade_num


def check_global_loss(o, global_stop_loss):
	margin = o.get_cash_details(verbose=True)
	pnl = margin['utilised']['m2m_realised']
	
	if pnl < 0:
		cash = margin['available']
		total_cash = cash['opening_balance'] + cash['intraday_payin'] + cash['collateral']
		
		return False if abs(pnl) > (total_cash * global_stop_loss) else margin['net']
	
	return margin['net']


def get_total_cash(o):
	margin = o.get_cash_details(verbose=True)
	cash = margin['available']
	
	return cash['opening_balance'] + cash['intraday_payin'] + cash['collateral']


def get_oms_session():
	s3 = boto3.client('s3')
	file_obj = s3.get_object(Bucket=os.environ['SESSION_BUCKET'], Key='session.json')
	
	return file_obj['Body'].read().decode('utf-8')


def ticker_transform(ticker):
	months = {'1': 'JAN', '2': 'FEB', '3': 'MAR', '4': 'APR', '5': 'MAY', '6': 'JUN', '7': 'JUL', '8': 'AUG',
	          '9': 'SEP', 'O': 'OCT', 'N': 'NOV', 'D': 'DEC'}
	date_modifier = {'01': 'st', '02': 'nd', '03': 'rd', '21': 'st', '22': 'nd', '23': 'rd'}
	# NIFTY 24 22221150CE
	try:
		cur_year_start_index = re.search(r'\d\d', ticker).start()
	except AttributeError:
		# if the ticker is just a stock ticker and not a future or option
		return ticker
	
	cur_year = ticker[cur_year_start_index: cur_year_start_index + 2]
	_ticker, data = ticker[:cur_year_start_index], ticker[cur_year_start_index + 2:]
	
	if any([x in data for x in months.values()]):
		month, date, strike = data[:3].title(), '', data[3:]
		date_modified = ''
	else:
		month, date, strike = months[data[0]].title(), data[1:3], data[3:]
		date_modified = date_modifier.get(date, 'th')
	
	return f'{_ticker} {date}{date_modified} {month} 20{cur_year} {strike}'


def round_down(x, nearest_num=50):
	return x if x % nearest_num == 0 else x - (x % nearest_num)


def round_up(x, nearest_num=50):
	return x if x % nearest_num == 0 else x + (nearest_num - (x % nearest_num))



import decimal
import gzip
import json
import pickle
import shutil


class DecimalEncoder(json.JSONEncoder):
	def default(self, o):
		if isinstance(o, decimal.Decimal):
			return str(o)
		
		return super().default(o)


def uncompress(zip_path, file_path):
	with gzip.open(zip_path, 'rb') as f_in:
		with open(file_path, 'wb') as f_out:
			shutil.copyfileobj(f_in, f_out)


def text_to_pickle(text, direc):
	with open(direc, 'wb') as f:
		pickle.dump(text, f)


def pickle_to_text(direc):
	with open(direc, 'rb') as f:
		return pickle.load(f)


def colourise(text, colour, decorate=None, end_='\n'):
	_c = {'pink': '\033[95m', 'blue': '\033[94m', 'green': '\033[92m', 'yellow': '\033[93m', 'grey': '\033[97m',
	      'cyan': '\033[96m', 'end': '\033[0m', 'red': '\033[91m', 'underline': '\033[4m', 'bold': '\033[1m'}
	colour, end = _c[colour], _c['end']
	
	if decorate is not None:
		print(f'{_c[decorate]}{colour}{text}{end}', end=end_)
	else:
		print(f'{colour}{text}{end}', end=end_)



import contextlib
import json
from datetime import datetime as dt

from quasi_utils.oms_utils import request, base, get_details, ticker_transform


class OMS:
	def __init__(self, config=None, data_dir=None, session=None):
		self.base_url = 'https://api.kite.trade'
		self.api_key, self.api_secret, self.access_token = get_details(config, data_dir)
		self.headers = {'X-Kite-Version': '3', 'User-Agent': 'Kiteconnect-python/4.2.0',
		                'Authorization': f'token {self.api_key}:{self.access_token}'}
		self.session = session
	
	def ltp(self, tickers):
		if not isinstance(tickers, list):
			tickers = [tickers]
		tickers_ = {'i': tickers}
		
		res = request('GET', f'{self.base_url}/quote/ltp', data=tickers_, params=tickers_,
		              headers=self.headers)
		if res.get('status_code') == 403:
			return None
		
		return {ticker_.split(':')[-1]: ltp['last_price'] for ticker_, ltp in res['data'].items()}
	
	def quote(self, tickers):
		if not isinstance(tickers, list):
			tickers = [tickers]
		
		res = request('GET', f'{self.base_url}/quote', params={'i': tickers}, headers=self.headers)
		if res.get('status_code') == 403:
			return None
		
		return res['data']
	
	def get_user(self):
		res = request('GET', f'{self.base_url}/user/profile', headers=self.headers)
		if res.get('status_code') == 403:
			return None
		
		return res
	
	def get_cash_needed(self, ticker, action, price, qty, price_type, trade_type='MIS', variety='regular',
	                    exchange='NFO'):
		data = {'tradingsymbol': ticker, 'exchange': exchange, 'transaction_type': action, 'price': price,
		        'quantity': qty, 'variety': variety, 'order_type': price_type, 'product': trade_type}
		
		res = request('POST', f'{self.base_url}/margins/orders', headers=self.headers, _json=[data])
		if res.get('status_code') == 403:
			return None
		
		return res
	
	def get_cash_details(self, verbose=False):
		res = request('GET', f'{self.base_url}/user/margins/equity', headers=self.headers)
		
		if res.get('status_code') == 403:
			return None
		
		data = res['data']
		
		if not data:
			return None
		
		return data if verbose else {'cash': round(data['net'], 1), 'pnl': data['utilised']['m2m_realised']}
	
	def get_instruments(self, tickers=None, exchange='NSE'):
		res = request('GET', f'{self.base_url}/instruments/{exchange}', data=None, headers=self.headers)
		if type(res) is dict and res.get('status_code') == 403:
			return None
		
		res = res.decode('utf-8').strip().split('\n')
		header, rows = res[0], res[1:]
		
		# handle special case of sensex index
		data = {'SENSEX': {'token': '265', 'name': 'SENSEX', 'expiry': '', 'strike': '0',
		                   'ticker_type': 'EQ', 'segment': 'INDICES'}} if exchange == 'NSE' else {}
		
		if not tickers:
			for idx, row in enumerate(rows):
				row = row.split(',')
				expiry = dt.strptime(row[5], '%Y-%m-%d') if row[5] != '' else ''
				name = row[3][1:-1]
				
				if name == '' or any(x in name for x in ['-', 'ETF']):
					continue
				
				name = row[2] if exchange == 'NSE' else name
				data[row[2]] = {'token': row[0], 'name': name, 'expiry': expiry, 'strike': row[-6],
				                'ticker_type': row[-3], 'segment': row[-2]}
		
		compareFn = lambda x: (f"z{x[1]['name']}" if x[1]['name'] != 'NIFTY' else x[1]['name'], x[1]['expiry'])
		data = dict(sorted(data.items(), key=compareFn))
		
		for ticker, ticker_data in data.items():
			if ticker_data['expiry'] != '':
				ticker_data['expiry'] = dt.strftime(ticker_data['expiry'], '%d %b %Y')
			
			data[ticker] = '|'.join([val for val in data[ticker].values()])
		
		return data
	
	def get_orders(self, status='all'):
		res = request('GET', f'{self.base_url}/orders', data=None, headers=self.headers)
		if res.get('status_code') == 403:
			return None
		
		orders, open_orders, executed_orders, all_orders = res['data'], [], [], []
		values_dict = {'NRML': 'OVERNIGHT', 'MIS': 'INTRADAY', 'COMPLETE': 'COMPLETED'}
		
		for order in orders:
			temp_order = {'order_id': order['order_id'],
			              'status': values_dict.get(order['status']) or order['status'],
			              'time': order['order_timestamp'],
			              'variety': order['variety'],
			              'exchange': order['exchange'],
			              'ticker': ticker_transform(order['tradingsymbol']),
			              'token': order['instrument_token'],
			              'price_type': order['order_type'],
			              'action': order['transaction_type'].title(),
			              'trade_type': values_dict[order['product']],
			              'total_qty': order['quantity'],
			              'price': round(order['price'], 2),
			              'average_price': round(order['average_price'], 2),
			              'filled_qty': order['filled_quantity'],
			              'pending_qty': order['pending_quantity']}
			
			if order['meta']:
				temp_order['total_qty'] = order['meta']['iceberg']['total_quantity']
			
			if temp_order['status'] in ['OPEN', 'AMO REQ RECEIVED']:
				open_orders.append(temp_order)
			else:
				executed_orders.append(temp_order)
				
			all_orders.append(temp_order)
		
		if status == 'open':
			return open_orders[::-1]
		elif status == 'executed':
			return executed_orders[::-1]
		else:
			return all_orders[::-1]
	
	def get_positions(self, only_open=False):
		res = request('GET', f'{self.base_url}/portfolio/positions', data=None, headers=self.headers)
		if res.get('status_code') == 403:
			return None
		
		positions = res['data']['net']
		
		final_positions = []
		for position in positions:
			temp_position = {'ticker': ticker_transform(position['tradingsymbol']),
			                 'exchange': position['exchange'],
			                 'token': position['instrument_token'],
			                 'trade_type': 'OVERNIGHT' if position['product'] == 'NRML' else 'INTRADAY',
			                 'qty': position['quantity'],
			                 'overnight_qty': position['overnight_quantity'],
			                 'avg_price': position['average_price'],
			                 'pnl': position['pnl'],
			                 'm2m': position['m2m'],
			                 'unrealised': position['unrealised'],
			                 'realised': position['realised'],
			                 'day_buy_price': position['day_buy_price'],
			                 'day_sell_price': position['day_sell_price']}
			final_positions.append(temp_position)
		
		return [position for position in final_positions if position['quantity']] if only_open else final_positions
	
	def place_order(self, ticker, action, price, qty, price_type, trade_type='MIS', variety='regular', exchange='NFO',
	                trigger_price=None, iceberg_legs=None, iceberg_quantity=None):
		data = {'tradingsymbol': ticker, 'exchange': exchange, 'transaction_type': action, 'price': price,
		        'quantity': qty, 'variety': variety, 'order_type': price_type, 'product': trade_type,
		        'trigger_price': trigger_price, 'iceberg_legs': iceberg_legs, 'iceberg_quantity': iceberg_quantity}
		
		res = request('POST', f'{self.base_url}/orders/{variety}', data=data, headers=self.headers)
		if res.get('status_code') == 403:
			return None
		
		return res
	
	def cancel_order(self, order_ids, variety='regular'):
		res = {}
		
		order_ids = [order_ids] if not isinstance(order_ids, list) else order_ids
		for _id in order_ids:
			res = request('DELETE', f'{self.base_url}/orders/{variety}/{_id}',
			              headers=self.headers, data=None)
			if res.get('status_code') == 403:
				return None
			
			res[_id] = res
		
		return res
	
	def modify_order(self, order_id=None, price=None, qty=None, price_type=None, variety='regular'):
		data = {}
		if price:
			data['price'] = price
		if qty:
			data['quantity'] = qty
		if price_type:
			data['order_type'] = price_type
		
		res = request('PUT', f'{self.base_url}/orders/{variety}/{order_id}',
		              data=data, headers=self.headers)
		if res.get('status_code') == 403:
			return None
		
		return res
	
	def place_gtt(self, buy_price, trade_type, qty, thresh=None, ticker=None, prefix=None, strike=None, exchange='NFO'):
		base_price, ticker = base(buy_price * thresh) if thresh else base(buy_price), ticker or f'{prefix}{strike}'
		
		condition = {'exchange': exchange, 'tradingsymbol': ticker, 'trigger_values': [base_price],
		             'last_price': self.ltp(f'{exchange}:{ticker}')[ticker]}
		orders = [{'exchange': exchange, 'tradingsymbol': ticker, 'transaction_type': 'SELL', 'quantity': qty,
		           'order_type': 'LIMIT', 'product': trade_type, 'price': base(base_price * 0.98)}]
		data = {'condition': json.dumps(condition), 'orders': json.dumps(orders), 'type': 'single'}
		
		res = request('POST', f'{self.base_url}/gtt/triggers', data=data, headers=self.headers)
		if res.get('status_code') == 403:
			return None
		
		return res
	
	def delete_gtt_orders(self, ids):
		if ids == 'all':
			for order in self.fetch_gtt_orders():
				res = request('DELETE', f'{self.base_url}/gtt/triggers/{order["id"]}',
				              headers=self.headers, data=None)
				if res.get('status_code') == 403:
					return None
		else:
			if not isinstance(ids, list):
				ids = [ids]
			
			for id_ in ids:
				res = request('DELETE', f'{self.base_url}/gtt/triggers/{id_}',
				              headers=self.headers, data=None)
				if res.get('status_code') == 403:
					return None
	
	def fetch_gtt_orders(self):
		res = request('GET', f'{self.base_url}/gtt/triggers', headers=self.headers, data=None)
		if res.get('status_code') == 403:
			return None
		
		return res['data']


if __name__ == '__main__':
	obj = OMS(config='zerodha', data_dir='../../oms/oms/execute/data')
	# print(obj.ltp(['NSE:RELIANCE']))
	# print(obj.get_cash_details(verbose=True))
	# print(json.dumps(obj.get_orders(mould=True), indent=2))
	# print(obj.cancel_order(order_ids='231004203428841', variety='amo'))
	# print(obj.modify_order(order_id='231004203471297', price=12.9, qty=500, price_type='LIMIT', variety='amo'))
	# print(obj.get_instruments())
	# print(obj.get_user()['data'])
	# print(json.dumps(obj.get_positions(), indent=2))
	print(json.dumps(obj.get_orders(status='executed'), indent=2))
